# This file is placed in the Public Domain.


"stub"


import setuptools


if __name__ == "__main__":
    setuptools.setup(scripts=[
        'bin/otpcr',
        'bin/otpcrc',
        'bin/otpcrd',
        'bin/otpcrs'
    ])
